#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <linux/kernel.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <linux/pci.h>

#include "dpci8255multi.h"

#include <asm/uaccess.h>

/* I defined some macro for device_open function */
#define TYPE(dev) (MINOR(dev) >> 4) 
#define NUM(dev) (MINOR(dev) & 0x0f)

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

#ifndef DECISION_VENDOR_ID
#define DECISION_VENDER_ID 0x6666
#endif

#ifndef DECISION_DEVICE_ID
#define DECISION_DEVICE_ID 0x0102
#endif

static int dpci8255multi_cards;
static char dpci8255multi_rev[2];
static int dpci8255multi_direction[MAXCARD][6];
static int dpci8255multi_counterbcd[MAXCARD][3];
static int dpci8255multi_countermode[MAXCARD][3];
int dpci8255multi_Base_Addr[MAXCARD];
/*---------------------------------------------------------------------*/
int dpci8255multi_open(struct inode *inoder,struct file *flip)
{
/* int type= TYPE(inoder->i_rdev);*/

 int num=NUM(inoder->i_rdev);
 if(num>=dpci8255multi_cards) return -ENODEV;

 return 0;
}
/*---------------------------------------------------------------------*/
static int dpci8255multi_release(struct inode *inoder,struct file *flip)
{
	return 0;
}
/*---------------------------------------------------------------------*/
static ssize_t dpci8255multi_read(struct file *flip,char *buf,size_t count,loff_t *ptr)
{
 int countt=0;

  int num=MINOR(flip->f_dentry->d_inode->i_rdev);
  if(ptr!=&flip->f_pos)  /* can't seek on this device */
   return -ESPIPE;
 
 if(num>=dpci8255multi_cards)
   return -EINVAL;
 buf[0]=readb(dpci8255multi_Base_Addr[0]);
 countt=1;
           
 return countt;
}
/*---------------------------------------------------------------------*/
static ssize_t dpci8255multi_write(struct file *flip,const char *buf,size_t count,loff_t *ptr)
{
 if(ptr!=&flip->f_pos)  /* can't seek on this device */
  return -ESPIPE;
 
 return 0;       /* do nothing */
}
/*---------------------------------------------------------------------*/
int dpci8255multi_Myread(int num,unsigned char *buf,int count)
{
  buf[2]=inb(dpci8255multi_Base_Addr[num]+(int)buf[1]);
  return 3;
}
/*---------------------------------------------------------------------*/
void dpci8255multi_Mywrite(int num,unsigned char *buf,int count)
{
   outb(buf[2],dpci8255multi_Base_Addr[num]+(int)buf[1]);
}
/*---------------------------------------------------------------------*/
int dpci8255multi_ioctl(
    struct inode *inode,
    struct file *file,
    unsigned int ioctl_num,/* The number of the ioctl */
    unsigned long ioctl_param) /* The parameter to it */
{
  unsigned char param[4];
  char *temp=(char *)ioctl_param;
  char ch1=0;
  int data;
  int num=NUM(inode->i_rdev);
  switch (ioctl_num) 
   {

    case DPCI8255MULTI_SET_REGISTER:
      ch1=(char)((int)temp);
        dpci8255multi_rev[1]=ch1;
      break;

    case DPCI8255MULTI_WRITE_PORT:
      param[0]=dpci8255multi_rev[0];
      param[1]=dpci8255multi_rev[1];
      param[2]=(char)((int)temp);
      dpci8255multi_Mywrite(num,param,4);
      break;

    case DPCI8255MULTI_READ_PORT:
      param[0]=dpci8255multi_rev[0];
      param[1]=dpci8255multi_rev[1];
      dpci8255multi_Myread(num,param,4);
      put_user(param[2],temp);
      break;

    case DPCI8255MULTI_READ_COUNTER:
      param[0]=dpci8255multi_rev[0];
      param[1]=dpci8255multi_rev[1];
      dpci8255multi_Myread(num,param,4);
      put_user(param[2],temp);
      break; 
      
    case DPCI8255MULTI_WRITE_COUNTER:
      data=(int)ioctl_param;
      param[0]=dpci8255multi_rev[0];
      param[1]=dpci8255multi_rev[1];
      param[2]=(char)data;
      param[3]=(char)(data>>8);
      dpci8255multi_Mywrite(num,param,4);
      break;

    case DPCI8255MULTI_SET_CONTROLWORD:
      param[0]=10;
      param[1]=dpci8255multi_rev[1];
      param[2]=(char)((int)temp);
      dpci8255multi_Mywrite(num,param,4);
      break;      
            
    case DPCI8255MULTI_GET_CONTROLWORD:
      param[0]=10;
      param[1]=dpci8255multi_rev[1];
      dpci8255multi_Myread(num,param,4);
      put_user(param[2],temp);
      break;
     default:
             break;
  }

  return 0;
}
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
static struct file_operations dpci8255multi_fops={
read: dpci8255multi_read,
write: dpci8255multi_write,
ioctl: dpci8255multi_ioctl ,
open: dpci8255multi_open,
release: dpci8255multi_release,
};
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
MODULE_AUTHOR("Delta");
MODULE_DESCRIPTION("Decision PCI Multi 8255 Driver");
MODULE_LICENSE("GPL");
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/

static int __init dpci8255multi_init(void)
{
 int index=0;
 struct pci_dev *dev = NULL;
 static int dev_count=0;
 int i; 

 dev_count++;
 if(dev_count>1)
  {
   printk(KERN_ERR"This drive only can open one time.\n");
   return -ENODEV;
  }

#if 0
#ifdef CONFIG_PCI
 if (!pci_present())   /* No PCI bus in this machine! */
#endif
 return -ENODEV;
#endif

 dpci8255multi_cards=0;
 while (index < MAXCARD && /* dev == NULL && */
        (dev = pci_find_device(DECISION_VENDER_ID, DECISION_DEVICE_ID,dev))) 
 {
   dpci8255multi_Base_Addr[index]=pci_resource_start(dev,3) & PCI_BASE_ADDRESS_IO_MASK;   
   printk(KERN_ERR "Decision PCI Multi 8255:I got a base %x\n",dpci8255multi_Base_Addr[index]); 
   if(check_region(dpci8255multi_Base_Addr[index],REGIONLENGTH)>=0)
    request_region(dpci8255multi_Base_Addr[index],REGIONLENGTH,DEVICE_NAME);
   
   for(i=0;i<6;i++)
    dpci8255multi_direction[index][i]=0;  // default is output
   for(i=0;i<3;i++)
    {
     dpci8255multi_countermode[index][i]=3;
     dpci8255multi_counterbcd[index][i]=0;
    } 
   index++;
  }

  if((dev==NULL)&&(index<1))
   {
    printk("\nI can't find any Decision PCI Multi 8255 card here!!\n");
    return -ENODEV;
   }
 printk(KERN_ERR "\nDecision PCI Multi 8255: Total %d cards that I catched!\n",index);
 dpci8255multi_cards=index;
 index=register_chrdev(dpci8255multi_major,DEVICE_NAME,&dpci8255multi_fops);
 if(index<0)
  {
  for(i=0;i<dpci8255multi_cards;i++)
   release_region(dpci8255multi_Base_Addr[i],REGIONLENGTH);
  return index;
  }
 dpci8255multi_rev[0]=0;
 dpci8255multi_rev[1]=0;
 printk("Decision PCI Multi 8255:Register Successful\n");  
 return 0;

}
/*---------------------------------------------------------------------*/
static void __exit dpci8255multi_exit(void)
{
  int i;
  for(i=0;i<dpci8255multi_cards;i++)
   release_region(dpci8255multi_Base_Addr[i],REGIONLENGTH); 
  unregister_chrdev(dpci8255multi_major,DEVICE_NAME);
  printk("Decision PCI Multi 8255:Unregister Successful!!\n");  
}
/*---------------------------------------------------------------------*/
module_init(dpci8255multi_init);
module_exit(dpci8255multi_exit);
